@@ -24,6 +24,8 @@
 #endif
 
 #include <errno.h>
+#include <string.h>
+
 #include <gdbus.h>
 
 #include "connman.h"
@@ -36,20 +38,42 @@ struct connman_device {
 	connman_bool_t powered;
 	connman_bool_t carrier;
 	connman_bool_t scanning;
+	connman_bool_t disconnected;
 	connman_uint8_t priority;
+	connman_uint16_t scan_interval;
 	char *name;
 	char *node;
 	char *interface;
 	unsigned int connections;
+	guint scan_timeout;
 
 	struct connman_device_driver *driver;
 	void *driver_data;
 
 	connman_bool_t registered;
 
+	char *last_network;
+	struct connman_network *network;
 	GHashTable *networks;
 };
 
+static gboolean device_scan_trigger(gpointer user_data)
+{
+	struct connman_device *device = user_data;
+
+	DBG("device %p", device);
+
+	if (device->driver == NULL) {
+		device->scan_timeout = 0;
+		return FALSE;
+	}
+
+	if (device->driver->scan)
+		device->driver->scan(device);
+
+	return TRUE;
+}
+
 static const char *type2description(enum connman_device_type type)
 {
 	switch (type) {
@@ -213,6 +237,15 @@ static int set_policy(DBusConnection *connection,
 	return 0;
 }
 
+static void append_path(gpointer key, gpointer value, gpointer user_data)
+{
+	struct connman_element *element = value;
+	DBusMessageIter *iter = user_data;
+
+	dbus_message_iter_append_basic(iter, DBUS_TYPE_OBJECT_PATH,
+							&element->path);
+}
+
 static void append_networks(struct connman_device *device,
 						DBusMessageIter *entry)
 {
@@ -227,8 +260,7 @@ static void append_networks(struct connman_device *device,
 
 	dbus_message_iter_open_container(&value, DBUS_TYPE_ARRAY,
 				DBUS_TYPE_OBJECT_PATH_AS_STRING, &iter);
-	__connman_element_list((struct connman_element *) device,
-					CONNMAN_ELEMENT_TYPE_NETWORK, &iter);
+	g_hash_table_foreach(device->networks, append_path, &iter);
 	dbus_message_iter_close_container(&value, &iter);
 
 	dbus_message_iter_close_container(entry, &value);
@@ -295,10 +327,16 @@ static DBusMessage *get_properties(DBusConnection *conn,
 
 	switch (device->mode) {
 	case CONNMAN_DEVICE_MODE_UNKNOWN:
+		break;
 	case CONNMAN_DEVICE_MODE_TRANSPORT_IP:
+		__connman_element_append_ipv4(&device->element, &dict);
 		break;
 	case CONNMAN_DEVICE_MODE_NETWORK_SINGLE:
 	case CONNMAN_DEVICE_MODE_NETWORK_MULTIPLE:
+		if (device->scan_interval > 0)
+			connman_dbus_dict_append_variant(&dict, "ScanInterval",
+				DBUS_TYPE_UINT16, &device->scan_interval);
+
 		dbus_message_iter_open_container(&dict, DBUS_TYPE_DICT_ENTRY,
 								NULL, &entry);
 		append_networks(device, &entry);
@@ -317,6 +355,7 @@ static DBusMessage *set_property(DBusConnection *conn,
 	struct connman_device *device = data;
 	DBusMessageIter iter, value;
 	const char *name;
+	int type;
 
 	DBG("conn %p", conn);
 
@@ -331,10 +370,15 @@ static DBusMessage *set_property(DBusConnection *conn,
 					CONNMAN_SECURITY_PRIVILEGE_MODIFY) < 0)
 		return __connman_error_permission_denied(msg);
 
+	type = dbus_message_iter_get_arg_type(&value);
+
 	if (g_str_equal(name, "Powered") == TRUE) {
 		connman_bool_t powered;
 		int err;
 
+		if (type != DBUS_TYPE_BOOLEAN)
+			return __connman_error_invalid_arguments(msg);
+
 		dbus_message_iter_get_basic(&value, &powered);
 
 		if (device->powered == powered)
@@ -348,6 +392,9 @@ static DBusMessage *set_property(DBusConnection *conn,
 		const char *str;
 		int err;
 
+		if (type != DBUS_TYPE_STRING)
+			return __connman_error_invalid_arguments(msg);
+
 		dbus_message_iter_get_basic(&value, &str);
 		policy = string2policy(str);
 		if (policy == CONNMAN_DEVICE_POLICY_UNKNOWN)
@@ -359,9 +406,52 @@ static DBusMessage *set_property(DBusConnection *conn,
 	} else if (g_str_equal(name, "Priority") == TRUE) {
 		connman_uint8_t priority;
 
+		if (type != DBUS_TYPE_BYTE)
+			return __connman_error_invalid_arguments(msg);
+
 		dbus_message_iter_get_basic(&value, &priority);
 
 		device->priority = priority;
+	} else if (g_str_equal(name, "ScanInterval") == TRUE) {
+		connman_uint16_t interval;
+
+		switch (device->mode) {
+		case CONNMAN_DEVICE_MODE_UNKNOWN:
+		case CONNMAN_DEVICE_MODE_TRANSPORT_IP:
+			return __connman_error_invalid_arguments(msg);
+		case CONNMAN_DEVICE_MODE_NETWORK_SINGLE:
+		case CONNMAN_DEVICE_MODE_NETWORK_MULTIPLE:
+			break;
+		}
+
+		if (type != DBUS_TYPE_UINT16)
+			return __connman_error_invalid_arguments(msg);
+
+		dbus_message_iter_get_basic(&value, &interval);
+
+		device->scan_interval = interval;
+
+		if (device->scan_timeout > 0) {
+			g_source_remove(device->scan_timeout);
+			device->scan_timeout = 0;
+		}
+
+		if (device->scan_interval > 0) {
+			guint interval = device->scan_interval;
+			device->scan_timeout = g_timeout_add_seconds(interval,
+						device_scan_trigger, device);
+		}
+	} else if (g_str_has_prefix(name, "IPv4") == TRUE) {
+		switch (device->mode) {
+		case CONNMAN_DEVICE_MODE_UNKNOWN:
+		case CONNMAN_DEVICE_MODE_NETWORK_SINGLE:
+		case CONNMAN_DEVICE_MODE_NETWORK_MULTIPLE:
+			return __connman_error_invalid_arguments(msg);
+		case CONNMAN_DEVICE_MODE_TRANSPORT_IP:
+			__connman_element_set_ipv4(&device->element,
+								name, &value);
+			break;
+		}
 	}
 
 	__connman_storage_save_device(device);
@@ -572,6 +662,9 @@ static void probe_driver(struct connman_element *element, gpointer user_data)
 	if (element->device->driver != NULL)
 		return;
 
+	if (driver->type != element->device->type)
+		return;
+
 	if (driver->probe(element->device) < 0)
 		return;
 
@@ -683,6 +776,8 @@ static void device_destruct(struct connman_element *element)
 	g_free(device->name);
 	g_free(device->interface);
 
+	g_free(device->last_network);
+
 	g_hash_table_destroy(device->networks);
 	device->networks = NULL;
 }
@@ -710,20 +805,21 @@ struct connman_device *connman_device_create(const char *node,
 
 	DBG("device %p", device);
 
-	device->element.refcount = 1;
+	__connman_element_initialize(&device->element);
 
 	device->element.name = g_strdup(node);
 	device->element.type = CONNMAN_ELEMENT_TYPE_DEVICE;
-	device->element.index = -1;
 
 	device->element.device = device;
 	device->element.destruct = device_destruct;
 
 	str = type2string(type);
 	if (str != NULL)
-		connman_element_add_static_property(&device->element,
+		connman_element_set_static_property(&device->element,
 					"Type", DBUS_TYPE_STRING, &str);
 
+	device->element.ipv4.method = CONNMAN_IPV4_METHOD_DHCP;
+
 	device->type   = type;
 	device->mode   = CONNMAN_DEVICE_MODE_UNKNOWN;
 	device->policy = CONNMAN_DEVICE_POLICY_AUTO;
@@ -732,22 +828,27 @@ struct connman_device *connman_device_create(const char *node,
 	case CONNMAN_DEVICE_TYPE_UNKNOWN:
 	case CONNMAN_DEVICE_TYPE_VENDOR:
 		device->priority = 0;
+		device->scan_interval = 0;
 		break;
 	case CONNMAN_DEVICE_TYPE_ETHERNET:
 	case CONNMAN_DEVICE_TYPE_WIFI:
 		device->priority = 100;
+		device->scan_interval = 300;
 		break;
 	case CONNMAN_DEVICE_TYPE_WIMAX:
 		device->priority = 20;
+		device->scan_interval = 0;
 		break;
 	case CONNMAN_DEVICE_TYPE_BLUETOOTH:
 		device->priority = 50;
+		device->scan_interval = 0;
 		break;
 	case CONNMAN_DEVICE_TYPE_HSO:
 	case CONNMAN_DEVICE_TYPE_NOZOMI:
 	case CONNMAN_DEVICE_TYPE_HUAWEI:
 	case CONNMAN_DEVICE_TYPE_NOVATEL:
 		device->priority = 60;
+		device->scan_interval = 0;
 		break;
 	}
 
@@ -881,6 +982,17 @@ void connman_device_set_mode(struct connman_device *device,
 	device->mode = mode;
 }
 
+/**
+ * connman_device_get_mode:
+ * @device: device structure
+ *
+ * Get network mode of device
+ */
+enum connman_device_mode connman_device_get_mode(struct connman_device *device)
+{
+	return device->mode;
+}
+
 /**
  * connman_device_set_powered:
  * @device: device structure
@@ -902,6 +1014,9 @@ int connman_device_set_powered(struct connman_device *device,
 
 	device->powered = powered;
 
+	if (device->registered == FALSE)
+		return 0;
+
 	signal = dbus_message_new_signal(device->element.path,
 				CONNMAN_DEVICE_INTERFACE, "PropertyChanged");
 	if (signal == NULL)
@@ -924,6 +1039,17 @@ int connman_device_set_powered(struct connman_device *device,
 	if (device->policy != CONNMAN_DEVICE_POLICY_AUTO)
 		return 0;
 
+	if (device->scan_timeout > 0) {
+		g_source_remove(device->scan_timeout);
+		device->scan_timeout = 0;
+	}
+
+	if (device->scan_interval > 0) {
+		guint interval = device->scan_interval;
+		device->scan_timeout = g_timeout_add_seconds(interval,
+					device_scan_trigger, device);
+	}
+
 	if (device->driver->scan)
 		device->driver->scan(device);
 
@@ -957,11 +1083,24 @@ int connman_device_set_carrier(struct connman_device *device,
 	device->carrier = carrier;
 
 	if (carrier == TRUE) {
+		enum connman_element_type type = CONNMAN_ELEMENT_TYPE_UNKNOWN;
 		struct connman_element *element;
 
+		switch (device->element.ipv4.method) {
+		case CONNMAN_IPV4_METHOD_UNKNOWN:
+		case CONNMAN_IPV4_METHOD_OFF:
+			return 0;
+		case CONNMAN_IPV4_METHOD_STATIC:
+			type = CONNMAN_ELEMENT_TYPE_IPV4;
+			break;
+		case CONNMAN_IPV4_METHOD_DHCP:
+			type = CONNMAN_ELEMENT_TYPE_DHCP;
+			break;
+		}
+
 		element = connman_element_create(NULL);
 		if (element != NULL) {
-			element->type  = CONNMAN_ELEMENT_TYPE_DHCP;
+			element->type  = type;
 			element->index = device->element.index;
 
 			if (connman_element_register(element,
@@ -974,6 +1113,25 @@ int connman_device_set_carrier(struct connman_device *device,
 	return 0;
 }
 
+void __connman_device_disconnect(struct connman_device *device)
+{
+	GHashTableIter iter;
+	gpointer key, value;
+
+	DBG("device %p", device);
+
+	connman_device_set_disconnected(device, TRUE);
+
+	g_hash_table_iter_init(&iter, device->networks);
+
+	while (g_hash_table_iter_next(&iter, &key, &value) == TRUE) {
+		struct connman_network *network = value;
+
+		if (connman_network_get_connected(network) == TRUE)
+			__connman_network_disconnect(network);
+	}
+}
+
 static void connect_known_network(struct connman_device *device)
 {
 	struct connman_network *network = NULL;
@@ -988,9 +1146,21 @@ static void connect_known_network(struct connman_device *device)
 	while (g_hash_table_iter_next(&iter, &key, &value) == TRUE) {
 		connman_uint8_t old_priority, new_priority;
 		connman_uint8_t old_strength, new_strength;
+		const char *name;
 
 		count++;
 
+		if (connman_network_get_available(value) == FALSE)
+			continue;
+
+		name = connman_network_get_string(value, "Name");
+		if (name != NULL && device->last_network != NULL) {
+			if (g_str_equal(name, device->last_network) == TRUE) {
+				network = value;
+				break;
+			}
+		}
+
 		if (connman_network_get_remember(value) == FALSE)
 			continue;
 
@@ -1030,6 +1200,34 @@ static void connect_known_network(struct connman_device *device)
 		device->driver->scan(device);
 }
 
+static void mark_network_unavailable(gpointer key, gpointer value,
+							gpointer user_data)
+{
+	struct connman_network *network = value;
+
+	if (connman_network_get_connected(network) == TRUE)
+		return;
+
+	connman_network_set_available(network, FALSE);
+}
+
+static gboolean remove_unavailable_network(gpointer key, gpointer value,
+							gpointer user_data)
+{
+	struct connman_network *network = value;
+
+	if (connman_network_get_connected(network) == TRUE)
+		return FALSE;
+
+	if (connman_network_get_remember(network) == TRUE)
+		return FALSE;
+
+	if (connman_network_get_available(network) == TRUE)
+		return FALSE;
+
+	return TRUE;
+}
+
 /**
  * connman_device_set_scanning:
  * @device: device structure
@@ -1070,12 +1268,32 @@ int connman_device_set_scanning(struct connman_device *device,
 
 	g_dbus_send_message(connection, signal);
 
-	if (scanning == TRUE)
+	if (scanning == TRUE) {
+		if (device->scan_timeout > 0) {
+			g_source_remove(device->scan_timeout);
+			device->scan_timeout = 0;
+		}
+
+		if (device->scan_interval > 0) {
+			guint interval = device->scan_interval;
+			device->scan_timeout = g_timeout_add_seconds(interval,
+						device_scan_trigger, device);
+		}
+
+		g_hash_table_foreach(device->networks,
+					mark_network_unavailable, NULL);
 		return 0;
+	}
+
+	g_hash_table_foreach_remove(device->networks,
+					remove_unavailable_network, NULL);
 
 	if (device->connections > 0)
 		return 0;
 
+	if (device->disconnected == TRUE)
+		return 0;
+
 	if (device->policy != CONNMAN_DEVICE_POLICY_AUTO)
 		return 0;
 
@@ -1084,6 +1302,35 @@ int connman_device_set_scanning(struct connman_device *device,
 	return 0;
 }
 
+/**
+ * connman_device_set_disconnected:
+ * @device: device structure
+ * @disconnected: disconnected state
+ *
+ * Change disconnected state of device (only for device with networks)
+ */
+int connman_device_set_disconnected(struct connman_device *device,
+						connman_bool_t disconnected)
+{
+	DBG("driver %p disconnected %d", device, disconnected);
+
+	switch (device->mode) {
+	case CONNMAN_DEVICE_MODE_UNKNOWN:
+	case CONNMAN_DEVICE_MODE_TRANSPORT_IP:
+		return -EINVAL;
+	case CONNMAN_DEVICE_MODE_NETWORK_SINGLE:
+	case CONNMAN_DEVICE_MODE_NETWORK_MULTIPLE:
+		break;
+	}
+
+	if (device->disconnected == disconnected)
+		return -EALREADY;
+
+	device->disconnected = disconnected;
+
+	return 0;
+}
+
 /**
  * connman_device_set_string:
  * @device: device structure
@@ -1128,10 +1375,10 @@ const char *connman_device_get_string(struct connman_device *device,
 	return NULL;
 }
 
-static void set_flightmode(struct connman_element *element, gpointer user_data)
+static void set_offlinemode(struct connman_element *element, gpointer user_data)
 {
 	struct connman_device *device = element->device;
-	connman_bool_t flightmode = GPOINTER_TO_UINT(user_data);
+	connman_bool_t offlinemode = GPOINTER_TO_UINT(user_data);
 	connman_bool_t powered;
 
 	DBG("element %p name %s", element, element->name);
@@ -1139,7 +1386,7 @@ static void set_flightmode(struct connman_element *element, gpointer user_data)
 	if (device == NULL)
 		return;
 
-	powered = (flightmode == TRUE) ? FALSE : TRUE;
+	powered = (offlinemode == TRUE) ? FALSE : TRUE;
 
 	if (device->powered == powered)
 		return;
@@ -1147,12 +1394,12 @@ static void set_flightmode(struct connman_element *element, gpointer user_data)
 	set_powered(device, powered);
 }
 
-int __connman_device_set_flightmode(connman_bool_t flightmode)
+int __connman_device_set_offlinemode(connman_bool_t offlinemode)
 {
-	DBG("flightmode %d", flightmode);
+	DBG("offlinmode %d", offlinemode);
 
 	__connman_element_foreach(NULL, CONNMAN_ELEMENT_TYPE_DEVICE,
-				set_flightmode, GUINT_TO_POINTER(flightmode));
+			set_offlinemode, GUINT_TO_POINTER(offlinemode));
 
 	return 0;
 }
@@ -1240,6 +1487,19 @@ int connman_device_remove_network(struct connman_device *device,
 	return 0;
 }
 
+void __connman_device_set_network(struct connman_device *device,
+					struct connman_network *network)
+{
+	const char *name;
+
+	if (network != NULL) {
+		name = connman_network_get_string(network, "Name");
+		device->last_network = g_strdup(name);
+	}
+
+	device->network = network;
+}
+
 /**
  * connman_device_register:
  * @device: device structure
@@ -1250,6 +1510,16 @@ int connman_device_register(struct connman_device *device)
 {
 	__connman_storage_load_device(device);
 
+	switch (device->mode) {
+	case CONNMAN_DEVICE_MODE_UNKNOWN:
+	case CONNMAN_DEVICE_MODE_TRANSPORT_IP:
+		break;
+	case CONNMAN_DEVICE_MODE_NETWORK_SINGLE:
+	case CONNMAN_DEVICE_MODE_NETWORK_MULTIPLE:
+		__connman_storage_init_network(device);
+		break;
+	}
+
 	return connman_element_register(&device->element, NULL);
 }
 
@@ -1398,6 +1668,24 @@ static int device_load(struct connman_device *device)
 	if (val > 0)
 		device->priority = val;
 
+	switch (device->mode) {
+	case CONNMAN_DEVICE_MODE_UNKNOWN:
+	case CONNMAN_DEVICE_MODE_TRANSPORT_IP:
+		break;
+	case CONNMAN_DEVICE_MODE_NETWORK_SINGLE:
+	case CONNMAN_DEVICE_MODE_NETWORK_MULTIPLE:
+		val = g_key_file_get_integer(keyfile, "Configuration",
+							"ScanInterval", NULL);
+		if (val > 0)
+			device->scan_interval = val;
+		break;
+	}
+
+	str = g_key_file_get_string(keyfile, "Configuration",
+							"LastNetwork", NULL);
+	if (str != NULL)
+		device->last_network = str;
+
 	g_key_file_free(keyfile);
 
 	return 0;
@@ -1439,6 +1727,22 @@ update:
 		g_key_file_set_integer(keyfile, "Configuration",
 						"Priority", device->priority);
 
+	switch (device->mode) {
+	case CONNMAN_DEVICE_MODE_UNKNOWN:
+	case CONNMAN_DEVICE_MODE_TRANSPORT_IP:
+		break;
+	case CONNMAN_DEVICE_MODE_NETWORK_SINGLE:
+	case CONNMAN_DEVICE_MODE_NETWORK_MULTIPLE:
+		if (device->scan_interval > 0)
+			g_key_file_set_integer(keyfile, "Configuration",
+					"ScanInterval", device->scan_interval);
+		break;
+	}
+
+	if (device->last_network != NULL)
+		g_key_file_set_string(keyfile, "Configuration",
+					"LastNetwork", device->last_network);
+
 	data = g_key_file_to_data(keyfile, &length, NULL);
 
 	g_file_set_contents(pathname, data, length, NULL);
@@ -1450,6 +1754,9 @@ done:
 
 	g_free(pathname);
 
+	if (device->network != NULL)
+		__connman_storage_save_network(device->network);
+
 	return 0;
 }
 
